"""
`tz_curves` module
==================

"""

# Import libraries
import math as m
import numpy as np
from numba import njit, prange
from random import random

# SPRING FUNCTIONS --------------------------------------------

# API clay function
@njit(cache=True)
def api_clay(
    sig: float,
    Su: float,
    D: float,
    residual: float = 0.9,
    tensile_factor: float = 1.0,
    output_length: int = 15,
):
    """
    Creates the API clay t-z curve from relevant input.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    Su : float
        Undrained shear strength [unit: kPa]
    D: float
        Pile diameter [unit: m]
    residual: float
        residual strength after peak strength, according to API-RP-2A,
        this value is between 0.7 and 0.9, default to 0.9
    tensile_factor: float
        strength factor for negative values of the curve
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]
    """
    # cannot have less than 15
    if output_length < 15:
        output_length = 15

    # important variables
    if sig == 0.0:
        psi = Su / 0.001
    else:
        psi = Su / sig

    if psi > 1.0:
        alpha = min(0.5 * psi ** (-0.25), 1.0)
    else:
        alpha = min(0.5 * psi ** (-0.5), 1.0)

    # Unit skin friction [kPa]
    f = alpha * Su

    # piecewise function
    zlist = [0.0, 0.0016, 0.0031, 0.0057, 0.0080, 0.0100, 0.0200, 0.0300]
    tlist = [0.0, 0.3, 0.5, 0.75, 0.90, 1.00, residual, residual]

    # determine z vector
    z = np.array(zlist, dtype=np.float32) * D
    z = np.concatenate((-z[-1:0:-1], z))
    # define t vector
    t = np.array(tlist, dtype=np.float32) * f
    t = np.concatenate((-tensile_factor * t[-1:0:-1], t))

    add_values = output_length - len(z)
    add_z_values = np.zeros((add_values), dtype=np.float32)
    add_t_values = np.zeros((add_values), dtype=np.float32)

    for i in range(add_values):
        add_z_values[i] = (0.02 + random() * 0.01) * D
        add_t_values[i] = residual * f

    z = np.append(z, add_z_values)
    t = np.append(t, add_t_values)

    z = np.sort(z)
    z_id_sorted = np.argsort(z)

    t = t[z_id_sorted]

    return t, z


# API sand function
@njit(cache=True)
def api_sand(
    sig: float,
    delta: float,
    K: float = 0.8,
    tensile_factor: float = 1.0,
    output_length: int = 7,
):
    """
    Creates the API sand t-z curve from relevant input.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    delta: float
        interface friction angle [unit: degrees]
    K: float
        coefficient of lateral pressure (0.8 for open-ended piles and 1.0 for cloased-ended)
    tensile_factor: float
        strength factor for negative values of the curve
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 7, by default 7

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]
    """
    # cannot have less than 7
    if output_length < 7:
        output_length = 7

    # important variables
    delta_table = np.array([0, 15, 20, 25, 30, 35, 100], dtype=np.float32)
    fs_max_table = np.array([47.8, 47.8, 67, 81.3, 95.7, 114.8, 114.8], dtype=np.float32)

    # limit unit skin friction according to API ref page 59
    fs_max = np.interp(delta, delta_table, fs_max_table)

    # Unit skin friction [kPa]
    f = min(fs_max, K * sig * m.tan(delta * m.pi / 180.0))

    # piecewise function
    zlist = [0.0, 0.0254, 0.03, 0.04]
    tlist = [0.0, 1.0, 1.0, 1.0]

    # determine z vector
    z = np.array(zlist, dtype=np.float32)
    z = np.concatenate((-z[-1:0:-1], z))
    # define t vector
    t = np.array(tlist, dtype=np.float32) * f
    t = np.concatenate((-tensile_factor * t[-1:0:-1], t))

    add_values = output_length - len(z)
    add_z_values = np.zeros((add_values), dtype=np.float32)
    add_t_values = np.zeros((add_values), dtype=np.float32)

    for i in range(add_values):
        add_z_values[i] = 0.03 + random() * 0.01
        add_t_values[i] = f

    z = np.append(z, add_z_values)
    t = np.append(t, add_t_values)

    z = np.sort(z)
    z_id_sorted = np.argsort(z)

    t = t[z_id_sorted]

    return t, z



# misc functions


import random
import numpy as np
import pandas as pd
import math as m

from numba import njit
import matplotlib.colors as mcolors


def from_list2x_parse_top_bottom(var):
    """provide top and bottom values of layer based on float or list inputs"""
    if isinstance(var, float) or isinstance(var, int):
        top = var
        bottom = var
    elif isinstance(var, list) and len(var) == 1:
        top = var[0]
        bottom = var[0]
    elif isinstance(var, list) and len(var) == 2:
        top = var[0]
        bottom = var[1]
    else:
        print("Soil Layer variable is not a float nor a list")
        raise TypeError

    return top, bottom


def var_to_str(var):
    if isinstance(var, float) or isinstance(var, int):
        var_print = var
    elif isinstance(var, list):
        var_print = "-".join(str(v) for v in var)
    else:
        raise ValueError("not a float nor list")
    return var_print


def generate_color_string(kind=None):
    if kind is None:
        colors = list(mcolors.CSS4_COLORS.values())
    elif kind == "earth":
        colors = [
            "#B4A390",
            "#927E75",
            "#796363",
            "#99A885",
            "#D4DBBA",
            "#FDEDCF",
            "#EEEAC3",
            "#F5D498",
            "#ECB992",
            "#DDA175",
            "#AB7B5E",
            "#8F6854",
        ]
    return colors[random.randint(0, len(colors) - 1)]


def repeat_inner(arr):
    arr = arr.reshape(-1, 1)

    arr_inner = arr[1:-1]
    arr_inner = np.tile(arr_inner, (2)).reshape(-1)

    return np.hstack([arr[0], arr_inner, arr[-1]])


def get_springs(springs: np.ndarray, elevations: np.ndarray, kind: str) -> pd.DataFrame:
    """
    Returns soil springs created for the given model in one DataFrame.

    Parameters
    ----------
    springs : ndarray dim[nelem,2,2,spring_dim]
        Springs at top and bottom of element
    elevations : ndarray
        self.nodes_coordinates["x [m]"].values
    kind : str
        type of spring to extract. one of ["p-y", "m-t", "Hb-y", "Mb-t", "t-z"]

    Returns
    -------
    pd.DataFrame
        Soil springs
    """

    def springs_to_df(springs: np.ndarray, elevations: np.ndarray, flag) -> pd.DataFrame:

        spring_dim = springs.shape[-1]
        nelem = springs.shape[0]

        column_values_spring = [f"VAL {i}" for i in range(spring_dim)]

        id = np.repeat(np.arange(nelem), 4)
        x = np.repeat(repeat_inner(elevations), 2)

        if len(x) > 2:
            t_b = ["top", "top", "bottom", "bottom"] * int(nelem)

            df = pd.DataFrame(
                data={
                    "Element no.": id,
                    "Position": t_b,
                    "Elevation [m]": x,
                }
            )
        else:
            df = pd.DataFrame(
                data={
                    "Element no.": id,
                    "Elevation [m]": x,
                }
            )

        df["type"] = flag.split("-") * int(len(x) / 2)
        df[column_values_spring] = np.reshape(springs, (-1, spring_dim))

        return df

    return springs_to_df(springs, elevations, flag=kind)


@njit(cache=True)
def conic(
    x_u: float,
    n: float,
    k: float,
    y_u: float,
    output_length: int,
):
    # Create x vector with 10% extension
    x = np.array([0, 0.001, 0.005, 0.01]).astype(np.float32) * x_u
    x = np.append(x, np.linspace(0.02 * x_u, x_u, output_length - 5).astype(np.float32))
    x = np.append(x, 1.1 * x_u)

    a = 1 - 2 * n

    y = np.zeros((len(x)), dtype=np.float32)

    for i in range(len(x)):
        if abs(x[i] - x_u) < 1e-2:
            y[i] = y_u
        elif x[i] < x_u:
            b = 2 * n * x[i] / x_u - (1 - n) * (1 + x[i] * k / y_u)
            c = x[i] * (k / y_u) * (1 - n) - n * (x[i] ** 2 / x_u**2)

            y[i] = y_u * 2 * c / (-b + (b**2 - 4 * a * c) ** 0.5)
        else:
            y[i] = y_u

    return x, y
